#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <iterator>
#include <algorithm>
#include <functional>

using namespace std;

#include "xbyak/xbyak.h"

int befunge_op_g(class Befungec* t, int x, int y);
int befunge_op_p(class Befungec* t, int c, int x, int y);
void befunge_reload(class Befungec* t, int l, int x, int y);
int befunge_tmpbuf;

bool befunge_debugging = false;
bool befunge_non_selfmod = false;

class Befungec : Xbyak::CodeGenerator {
public:
    explicit Befungec(const char* file)
        : CodeGenerator(1024*1024), file_(file), initl_(0) {}
    Befungec(vector<string> lines, char l, int x, int y, int ss, int se)
        : lines_(lines), mx_(lines[0].size()), my_(lines_.size()),
          initl_(l), initx_(x), inity_(y),
          stack_start_(ss), stack_end_(se) {}

    void gen() {
        if (!initl_) load_();
        gen_();
    }

    void run() {
        const unsigned char* code = getCode();
        if (befunge_debugging) {
            printf("%p\n", code);
            ofstream ofs("out.bin");
            for (int i = 0; i < getSize(); i++) {
                ofs << code[i];
            }
            ofs.close();
        }

        ((void (*)())code)();
    }

private:
    // avoid undefined too far jmp
    void jmp(char* l) {
        Xbyak::CodeGenerator::jmp(l, T_NEAR);
    }

    char* label_(char l, int x, int y) {
        static char buf[256];
        sprintf(buf, "%c_%d_%d", l, x, y);
//        puts(buf);
        return buf;
    }
    char* left_label_(int x, int y) {
//        x = x > 0 ? x-1 : mx_-1;
        return label_('l', x-1, y);
    }
    char* right_label_(int x, int y) {
//        x = x < (int)mx_-1 ? x+1 : 0;
        return label_('r', x+1, y);
    }
    char* up_label_(int x, int y) {
//        y = y > 0 ? y-1 : my_-1;
        return label_('u', x, y-1);
    }
    char* down_label_(int x, int y) {
//        y = y < (int)my_-1 ? y+1 : 0;
        return label_('d', x, y+1);
    }

    void load_() {
        ifstream ifs(file_);
        string l;
        while (getline(ifs, l)) {
            // befunge97's comment
            if (l.size() > 0 && l[0] == '=') continue;
            lines_.push_back(l);
        }

        my_ = lines_.size();
        mx_ = 0;
        for (size_t i = 0; i < lines_.size(); i++) {
            size_t s = lines_[i].size();
            mx_ = max(s, mx_);
        }

        for (size_t i = 0; i < lines_.size(); i++) {
            size_t s = lines_[i].size();
            lines_[i].resize(mx_);
            for (; s < mx_; s++) {
                lines_[i][s] = ' ';
            }
        }
    }

    void gen_() {
        // this pointer
        mov(edi, (int)this);

        // save esp
        mov(eax, (int)&stack_start_);
        mov(ptr [eax], esp);

        if (initl_) {
            // restore stack
            int* ss = (int*)stack_start_;
            int* se = (int*)stack_end_;
            for (; ss >= se; ss--) {
//                printf("%d\n",*ss);
                push(*ss);
            }

            jmp(label_(initl_, initx_, inity_));
        }
        else {
            // init random generator
            rdtsc();
            mov(esi, eax);

            // for stack underflow
            for (int i = 0; i < 64; i++) {
                push(0);
            }
        }

        for (size_t i = 0; i < my_; i++) {
            gen_line_('r', 0, i, 1, 0);
            L(label_('r', mx_, i));
            jmp(label_('r', 0, i));

            gen_line_('l', mx_-1, i, -1, 0);
            L(label_('l', -1, i));
            jmp(label_('l', mx_-1, i));
        }
        for (size_t i = 0; i < mx_; i++) {
            gen_line_('d', i, 0, 0, 1);
            L(label_('d', i, my_));
            jmp(label_('d', i, 0));

            gen_line_('u', i, my_-1, 0, -1);
            L(label_('u', i, -1));
            jmp(label_('u', i, my_-1));
        }

    }

    void fpcall(void* fp) {
        mov(eax, (int)fp);
        call(eax);
    }

    void gen_line_(char l, size_t fx, size_t fy, size_t vx, size_t vy) {
        size_t x = fx;
        size_t y = fy;
        while (x < mx_ && y < my_) {
            L(label_(l, x, y));

            char op = lines_[y][x];

            if (befunge_debugging) {
                push(op);
                push(y);
                push(x);
                push((int)"(%d, %d) %d\n");
                fpcall((void*)&printf);
                pop(eax);
                pop(eax);
                pop(eax);
                pop(eax);
            }

            switch (op) {
            case '<':
                jmp(left_label_(x, y));
                break;
            case '>':
                jmp(right_label_(x, y));
                break;
            case '^':
                jmp(up_label_(x, y));
                break;
            case 'v':
                jmp(down_label_(x, y));
                break;
            case '_':
            {
                string s(string(label_(l, x, y))+"_L");
                pop(eax);
                cmp(eax, 0);
                je(s.c_str());
                jmp(left_label_(x, y));
                L(s.c_str());
                jmp(right_label_(x, y));
                break;
            }
            case '|':
            {
                string s(string(label_(l, x, y))+"_L");
                pop(eax);
                cmp(eax, 0);
                je(s.c_str());
                jmp(up_label_(x, y));
                L(s.c_str());
                jmp(down_label_(x, y));
                break;
            }
            case '?':
            {
                // random generator
                imul(esi, esi, 21303);
                lea(esi, ptr [esi+2531011]);
                mov(eax, esi);
                shr(eax, 16);
                and(eax, 3);

                string s(label_(l, x, y));
                cmp(eax, 0);
                je((s+"_l").c_str());
                cmp(eax, 1);
                je((s+"_r").c_str());
                cmp(eax, 2);
                je((s+"_u").c_str());
                jmp(down_label_(x, y));
                L((s+"_l").c_str());
                jmp(left_label_(x, y));
                L((s+"_r").c_str());
                jmp(right_label_(x, y));
                L((s+"_u").c_str());
                jmp(up_label_(x, y));

                break;
            }
            case ' ':
                break;
            case '#':
                x += vx;
                y += vy;
                break;
            case '@':
                push(0);
                fpcall((void*)&exit);
                break;
            case '"':
            {
                char c;
                size_t sx = x;
                size_t sy = y;
                sx += vx;
                sy += vy;
                while (sx < mx_ && sy < my_) {
                    c = lines_[sy][sx];
                    if (c == '"') break;
                    push(c);
                    sx += vx;
                    sy += vy;
                }
                sx += vx;
                sy += vy;
                if (sx < mx_ && sy < my_) {
                    jmp(label_(l, sx, sy));
                }
                else {
                    push((int)"Error: unmatched \"");
                    fpcall((void*)&puts);
                    push(1);
                    fpcall((void*)&exit);
                }
                break;
            }
            case '&':
                push((int)&befunge_tmpbuf);
                push((int)"%d");
                fpcall((void*)&scanf);
                pop(edx);
                pop(edx);
                mov(eax, (int)&befunge_tmpbuf);
                mov(eax, ptr [eax]);
                push(eax);
                break;
            case '~':
                fpcall((void*)&getchar);
                push(eax);
                break;
            case '.':
                push((int)"%d ");
                fpcall((void*)&printf);
                pop(eax);
                pop(eax);
                break;
            case ',':
                fpcall((void*)&putchar);
                pop(eax);
                break;
            case '+':
                pop(ebx);
                pop(eax);
                add(eax, ebx);
                push(eax);
                break;
            case '-':
                pop(ebx);
                pop(eax);
                sub(eax, ebx);
                push(eax);
                break;
            case '*':
                pop(ebx);
                pop(eax);
                imul(eax, ebx);
                push(eax);
                break;
            case '/':
                pop(ebx);
                pop(eax);
                mov(edx,0);
                idiv(ebx);
                push(eax);
                break;
            case '%':
                pop(ebx);
                pop(eax);
                mov(edx,0);
                idiv(ebx);
                push(edx);
                break;
            case '`':
                pop(eax);
                pop(ebx);
                cmp(eax, ebx);
                setl(al);
                push(eax);
                break;
            case '!':
                pop(eax);
                cmp(eax, 0);
                sete(al);
                push(eax);
                break;
            case ':':
                pop(eax);
                push(eax);
                push(eax);
                break;
            case '\\':
                pop(eax);
                pop(ebx);
                push(eax);
                push(ebx);
                break;
            case '$':
                pop(eax);
                break;
            case 'g':
                pop(edx);
                pop(eax);
                push(edx);
                push(eax);
                push(edi);
                fpcall((void*)&befunge_op_g);
                pop(edx);
                pop(edx);
                pop(edx);
                push(eax);
                break;
            case 'p':
            {
                pop(edx);
                pop(eax);
                pop(ecx);
                push(edx);
                push(eax);
                push(ecx);
                push(edi);
                fpcall((void*)&befunge_op_p);
                pop(edx);
                pop(edx);
                pop(edx);
                pop(edx);

                if (!befunge_non_selfmod) {
                    string tmplabel = string(label_(l, x, y))+"_p";
                    cmp(eax, 0);
                    je(tmplabel.c_str());

                    // save esp
                    mov(eax, (int)&stack_end_);
                    mov(ptr [eax], esp);

                    push(y+vy);
                    push(x+vx);
                    push(l);
                    push(edi);
                    fpcall((void*)&befunge_reload);

                    push((int)"Error: how can you come here!");
                    fpcall((void*)&puts);
                    push(1);
                    fpcall((void*)&exit);

                    L(tmplabel.c_str());
                }

                break;
            }
            default:
                if (op >= '0' && op <= '9') {
                    push(op-'0');
                }
            }

            // fill nops for self modifying
            x += vx;
            y += vy;
        }
    }

    int op_g(size_t x, size_t y) {
//        printf("%d %d\n", x, y);
        if (x >= mx_ || y >= my_) {
            printf("Error: 'g' out of bounds (%d, %d)\n", x, y);
            exit(1);
        }

        return lines_[y][x];
    }
    int op_p(size_t x, size_t y, int c) {
        if (x >= mx_ || y >= my_) {
            printf("Error: 'p' out of bounds (%d, %d)\n", x, y);
            exit(1);
        }

        string s("<>^v_|?#@\"&~.,+-*/%`!:\\$gp0123456789");
        int p = lines_[y][x];
        lines_[y][x] = c;
        if (p == c) return 0;
        if (s.find(c) != string::npos) return 1;
        if (s.find(p) != string::npos) return 1;
        return 0;
    }

    void reload(int l, int x, int y) {
        Befungec b(lines_, l, x, y, stack_start_, stack_end_);
        b.gen();
        b.run();
    }

    friend int befunge_op_g(class Befungec* t, int x, int y);
    friend int befunge_op_p(class Befungec* t, int c, int x, int y);
    friend void befunge_reload(class Befungec* t, int l, int x, int y);

private:
    const char* file_;
    vector<string> lines_;
    size_t mx_, my_;
    char initl_;
    size_t initx_, inity_;
    int stack_start_, stack_end_;
};

int befunge_op_g(class Befungec* t, int x, int y) {
    return t->op_g(x, y);
}
int befunge_op_p(class Befungec* t, int c, int x, int y) {
    return t->op_p(x, y, c);
}
void befunge_reload(class Befungec* t, int l, int x, int y) {
    t->reload(l, x, y);
}

void error(char* a) {
    printf("usage: %s <file>\n", a);
    exit(1);
}

int main(int argc, char* argv[]) {
    int ai = 1;

    if (argc < 2) error(argv[0]);

    while (true) {
        if (string(argv[ai]) == "-d") {
            befunge_debugging = true;
            ai++;
        }
        else if (string(argv[ai]) == "-n") {
            befunge_non_selfmod = true;
            ai++;
        }
        else break;
    }

    try {
        Befungec b = Befungec(argv[ai]);
        b.gen();
        b.run();
        printf("huh?\n");
    }
    catch (Xbyak::Error xe) {
        cout << Xbyak::ConvertErrorToString(xe) << endl;
    }

    return 1;
}
